package org.xbib.datastructures.multi;

import java.util.AbstractCollection;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Objects;
import java.util.Set;
import java.util.Spliterator;
import java.util.Spliterators;
import org.xbib.datastructures.api.Multimap;
import org.xbib.datastructures.api.Multiset;
import org.xbib.datastructures.api.SetMultimap;

/**
 * A skeleton {@code Multimap} implementation, not necessarily in terms of a {@code Map}.
 */
abstract class AbstractMultimap<K extends Object, V extends Object>
        implements Multimap<K, V> {
    @Override
    public boolean isEmpty() {
        return size() == 0;
    }

    @Override
    public boolean containsValue(Object value) {
        for (Collection<V> collection : asMap().values()) {
            if (collection.contains(value)) {
                return true;
            }
        }

        return false;
    }

    @Override
    public boolean containsEntry(Object key, Object value) {
        Collection<V> collection = asMap().get(key);
        return collection != null && collection.contains(value);
    }

    @Override
    public boolean remove(Object key, Object value) {
        Collection<V> collection = asMap().get(key);
        return collection != null && collection.remove(value);
    }

    @Override
    public boolean put(K key, V value) {
        return get(key).add(value);
    }

    @Override
    public boolean putAll(K key, Iterable<? extends V> values) {
        Objects.requireNonNull(values);
        // make sure we only call values.iterator() once
        // and we only call get(key) if values is nonempty
        if (values instanceof Collection<? extends V> valueCollection) {
            return !valueCollection.isEmpty() && get(key).addAll(valueCollection);
        } else {
            Iterator<? extends V> valueItr = values.iterator();
            return valueItr.hasNext() && addAll(get(key), valueItr);
        }
    }

    @Override
    public boolean putAll(Multimap<? extends K, ? extends V> multimap) {
        boolean changed = false;
        for (Entry<? extends K, ? extends V> entry : multimap.entries()) {
            changed |= put(entry.getKey(), entry.getValue());
        }
        return changed;
    }

    @Override
    public Collection<V> replaceValues(K key, Iterable<? extends V> values) {
        Objects.requireNonNull(values);
        Collection<V> result = removeAll(key);
        putAll(key, values);
        return result;
    }

    private transient Collection<Entry<K, V>> entries;

    @Override
    public Collection<Entry<K, V>> entries() {
        Collection<Entry<K, V>> result = entries;
        return (result == null) ? entries = createEntries() : result;
    }

    abstract Collection<Entry<K, V>> createEntries();

    class Entries extends MultimapsEntries<K, V> {
        @Override
        Multimap<K, V> multimap() {
            return AbstractMultimap.this;
        }

        @Override
        public Iterator<Entry<K, V>> iterator() {
            return entryIterator();
        }

        @Override
        public Spliterator<Entry<K, V>> spliterator() {
            return entrySpliterator();
        }
    }

    class EntrySet extends Entries implements Set<Entry<K, V>> {
        @Override
        public int hashCode() {
            return hashCodeImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            return equalsImpl(this, obj);
        }
    }

    abstract Iterator<Entry<K, V>> entryIterator();

    Spliterator<Entry<K, V>> entrySpliterator() {
        return Spliterators.spliterator(
                entryIterator(), size(), (this instanceof SetMultimap) ? Spliterator.DISTINCT : 0);
    }

    private transient Set<K> keySet;

    @Override
    public Set<K> keySet() {
        Set<K> result = keySet;
        return (result == null) ? keySet = createKeySet() : result;
    }

    abstract Set<K> createKeySet();

    private transient Multiset<K> keys;

    @Override
    public Multiset<K> keys() {
        Multiset<K> result = keys;
        return (result == null) ? keys = createKeys() : result;
    }

    abstract Multiset<K> createKeys();

    private transient Collection<V> values;

    @Override
    public Collection<V> values() {
        Collection<V> result = values;
        return (result == null) ? values = createValues() : result;
    }

    abstract Collection<V> createValues();

    class Values extends AbstractCollection<V> {
        @Override
        public Iterator<V> iterator() {
            return valueIterator();
        }

        @Override
        public Spliterator<V> spliterator() {
            return valueSpliterator();
        }

        @Override
        public int size() {
            return AbstractMultimap.this.size();
        }

        @Override
        public boolean contains(Object o) {
            return AbstractMultimap.this.containsValue(o);
        }

        @Override
        public void clear() {
            AbstractMultimap.this.clear();
        }
    }

    Iterator<V> valueIterator() {
        return valueIterator(entries().iterator());
    }

    Spliterator<V> valueSpliterator() {
        return Spliterators.spliterator(valueIterator(), size(), 0);
    }

    private transient Map<K, Collection<V>> asMap;

    @Override
    public Map<K, Collection<V>> asMap() {
        Map<K, Collection<V>> result = asMap;
        return (result == null) ? asMap = createAsMap() : result;
    }

    abstract Map<K, Collection<V>> createAsMap();

    // Comparison and hashing

    @Override
    public boolean equals(Object object) {
        return equalsImpl(this, object);
    }

    /**
     * Returns the hash code for this multimap.
     *
     * <p>The hash code of a multimap is defined as the hash code of the map view, as returned by
     * {@link Multimap#asMap}.
     *
     * @see Map#hashCode
     */
    @Override
    public int hashCode() {
        return asMap().hashCode();
    }

    /**
     * Returns a string representation of the multimap, generated by calling {@code toString} on the
     * map returned by {@link Multimap#asMap}.
     *
     * @return a string representation of the multimap
     */
    @Override
    public String toString() {
        return asMap().toString();
    }

    private static <K extends Object, V extends Object> Iterator<V> valueIterator(Iterator<Entry<K, V>> entryIterator) {
        return new TransformedIterator<Entry<K, V>, V>(entryIterator) {
            @Override
            V transform(Entry<K, V> entry) {
                return entry.getValue();
            }
        };
    }

    private static <T extends Object> boolean addAll(
            Collection<T> addTo, Iterator<? extends T> iterator) {
        Objects.requireNonNull(addTo);
        Objects.requireNonNull(iterator);
        boolean wasModified = false;
        while (iterator.hasNext()) {
            wasModified |= addTo.add(iterator.next());
        }
        return wasModified;
    }

    /** An implementation for {@link Set#hashCode()}. */
    static int hashCodeImpl(Set<?> s) {
        int hashCode = 0;
        for (Object o : s) {
            hashCode += o != null ? o.hashCode() : 0;

            hashCode = ~~hashCode;
            // Needed to deal with unusual integer overflow in GWT.
        }
        return hashCode;
    }

    /** An implementation for {@link Set#equals(Object)}. */
    private static boolean equalsImpl(Set<?> s, Object object) {
        if (s == object) {
            return true;
        }
        if (object instanceof Set) {
            Set<?> o = (Set<?>) object;

            try {
                return s.size() == o.size() && s.containsAll(o);
            } catch (NullPointerException | ClassCastException ignored) {
                return false;
            }
        }
        return false;
    }

    static boolean equalsImpl(Multimap<?, ?> multimap, Object object) {
        if (object == multimap) {
            return true;
        }
        if (object instanceof Multimap) {
            Multimap<?, ?> that = (Multimap<?, ?>) object;
            return multimap.asMap().equals(that.asMap());
        }
        return false;
    }
}
