package org.xbib.datastructures.multi;

import java.util.AbstractSet;
import java.util.Collection;
import java.util.Iterator;
import java.util.Objects;
import java.util.Set;
import org.xbib.datastructures.api.Multiset;

/**
 * {@link AbstractSet} substitute without the potentially-quadratic {@code removeAll}
 * implementation.
 */
abstract class ImprovedAbstractSet<E extends Object> extends AbstractSet<E> {
    @Override
    public boolean removeAll(Collection<?> c) {
        return removeAllImpl(this, c);
    }

    @Override
    public boolean retainAll(Collection<?> c) {
        return super.retainAll(Objects.requireNonNull(c));
    }

    private static boolean removeAllImpl(Set<?> set, Collection<?> collection) {
        Objects.requireNonNull(collection); // for GWT
        if (collection instanceof Multiset) {
            collection = ((Multiset<?>) collection).elementSet();
        }
        /*
         * AbstractSet.removeAll(List) has quadratic behavior if the list size
         * is just more than the set's size.  We augment the test by
         * assuming that sets have fast contains() performance, and other
         * collections don't.
         */
        if (collection instanceof Set && collection.size() > set.size()) {
            return removeAll(set.iterator(), collection);
        } else {
            return removeAllImpl(set, collection.iterator());
        }
    }

    /** Remove each element in an iterable from a set. */
    private static boolean removeAllImpl(Set<?> set, Iterator<?> iterator) {
        boolean changed = false;
        while (iterator.hasNext()) {
            changed |= set.remove(iterator.next());
        }
        return changed;
    }

    private static boolean removeAll(Iterator<?> removeFrom, Collection<?> elementsToRemove) {
        Objects.requireNonNull(elementsToRemove);
        boolean result = false;
        while (removeFrom.hasNext()) {
            if (elementsToRemove.contains(removeFrom.next())) {
                removeFrom.remove();
                result = true;
            }
        }
        return result;
    }
}
